package net.knarcraft.minecraftserverlauncher.utility;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.profile.ServerLauncherController;

import javax.swing.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.Scanner;

import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.downloadFile;
import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.getResourceAsScanner;
import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.readRemoteFile;

/**
 * A utility used for updating the software
 */
public final class Updater {

    private static final String updaterFile = Main.getApplicationWorkDirectory() + File.separator + "Updater.jar";
    private static final String updaterURL = "https://jenkins.knarcraft.net/job/KnarCraft/job/Jar-Updater/job/master/" +
            "lastSuccessfulBuild/artifact/target/jar-updater-1.0-SNAPSHOT.jar";
    private static final String targetFile = "minecraft-server-launcher.jar";

    /**
     * Gets the version channel and version of this software instance
     *
     * @return <p>An array of Channel, Version</p>
     * @throws FileNotFoundException <p>If unable to find or read currentversion.csv</p>
     */
    public static String[] getCurrentVersion() throws FileNotFoundException {
        Scanner file = getResourceAsScanner("currentversion.csv");
        if (!file.hasNextLine()) {
            throw new FileNotFoundException("File currentversion.csv is invalid");
        }
        String oldType = file.nextLine();
        if (!file.hasNextLine()) {
            throw new FileNotFoundException("File currentversion.csv is invalid");
        }
        String oldVer = file.nextLine();
        file.close();
        return new String[]{oldType, oldVer};
    }

    /**
     * Checks if a newer version is available
     *
     * @param updateURL     <p>The URL used for checking for updates</p>
     * @param updateChannel <p>The release channel to use</p>
     * @throws IOException <p>If the update data cannot be read</p>
     */
    public static void checkForUpdate(String updateURL, String updateChannel) throws IOException {
        String[] oldData = getCurrentVersion();
        String oldType = oldData[0];
        String oldVer = oldData[1];

        String data = readRemoteFile(updateURL);
        JsonObject jsonObject = new JsonParser().parse(data).getAsJsonObject();
        String latest = jsonObject.getAsJsonObject("latest").get(updateChannel).getAsString();

        if (!oldType.equals(updateChannel) || isVersionHigher(oldVer, latest)) {
            JsonArray versionList = jsonObject.getAsJsonArray("versions");
            String url = "";
            for (JsonElement elem : versionList) {
                JsonObject obj = elem.getAsJsonObject();
                String ver = obj.get("id").getAsString();
                String type = obj.get("type").getAsString();
                if (ver.equals(latest) && type.equals(updateChannel)) {
                    url = obj.get("url").getAsString();
                    break;
                }
            }
            if (!new File(updaterFile).exists()) {
                downloadFile(updaterURL, Paths.get(updaterFile));
            }

            int answer = JOptionPane.showConfirmDialog(null,
                    "An update is available. Do you want to update?", "Update available",
                    JOptionPane.YES_NO_OPTION
            );
            if (answer == JOptionPane.YES_NO_OPTION) {
                runUpdater(url);
            }
        }
    }

    /**
     * Decides whether one version number is higher than another
     *
     * @param oldVersion <p>The old version to check</p>
     * @param newVersion <p>The new version to check</p>
     * @return <p>True if the new version is higher than the old one</p>
     */
    public static boolean isVersionHigher(String oldVersion, String newVersion) {
        String[] oldVersionParts = oldVersion.split("\\.");
        String[] newVersionParts = newVersion.split("\\.");
        int versionLength = Math.max(oldVersionParts.length, newVersionParts.length);
        for (int i = 0; i < versionLength; i++) {
            int oldVersionNumber = oldVersionParts.length > i ? Integer.parseInt(oldVersionParts[i]) : 0;
            int newVersionNumber = newVersionParts.length > i ? Integer.parseInt(newVersionParts[i]) : 0;
            if (newVersionNumber != oldVersionNumber) {
                return newVersionNumber > oldVersionNumber;
            }
        }
        return false;
    }

    /**
     * Updates the software
     *
     * @param url <p>The URL of the new file to download</p>
     * @throws IOException <p>If unable to run the updater</p>
     */
    private static void runUpdater(String url) throws IOException {
        String javaCommand = ServerLauncherController.getInstance().getJavaCommand();
        ProcessBuilder builder;
        builder = new ProcessBuilder(javaCommand, "-jar", "Updater.jar", url, "yes", targetFile, "5", "nogui");
        builder.directory(new File(Main.getApplicationWorkDirectory()));
        builder.redirectErrorStream(true);
        builder.start();
        System.exit(1);
    }
}
